package ca.bcgsc.abyssexplorer.graph;

import java.util.LinkedList;
import java.util.List;

/**
 * Class to capture a k-mer overlap vertex.
 * 
 * @author Cydney Nielsen
 *
 */
public class Vertex {
	
	protected int id;
	// list of in and out edges
	protected List<Edge> incoming;
	protected List<Edge> outgoing;
	// need two lists - edge can connect to both poles (inverted repeats)
	protected List<Byte> incomingPoles;
	protected List<Byte> outgoingPoles;
	protected boolean isPalindrome;

	public Vertex(int id) {
		this.id = id;
		incoming = new LinkedList<Edge>();
		outgoing = new LinkedList<Edge>();
		incomingPoles = new LinkedList<Byte>();
		outgoingPoles = new LinkedList<Byte>();
		isPalindrome = false;
	}
	
	public int getId() {
		return id;
	}
	
	public void setIsPalindrome(boolean b) {
		isPalindrome = b;
	}
	
	public boolean isPalindrome() {
		return isPalindrome;
	}
	
	public boolean hasConsistentPoles() {
		
		// Under development...
		
		/*boolean cPoles = true;
		if (incomingPoles.size() == 0 && outgoingPoles.size() == 0) { 
			return true;
		}
		byte p;
		if (incomingPoles.size() != 0) {
			p = incomingPoles.get(0);
		} else {
			p = outgoingPoles.get(0);
		}
		// check for pole consistency
		for (byte b: incomingPoles) {
			if (b != p) {
				cPoles = false;
				break;
			}
		}
		if (cPoles) {
			for (byte b: outgoingPoles) {
				if (b != p) {
					cPoles = false;
					break;
				}
			}
		}
		return cPoles;*/
		return false;
	}
	
	public List<Edge> getIncoming() {
		if (isPalindrome) {
			return incoming;
		} else {
			List<Edge> tmpIncoming = new LinkedList<Edge>();
			for (Edge iEdge: incoming) {
				if (iEdge.getStrand() == 0) {
					tmpIncoming.add(iEdge);
				}
			}
			for (Edge oEdge: outgoing) {
				if (oEdge.getStrand() == 1) {
					tmpIncoming.add(oEdge);
				}
			}
			return tmpIncoming;
		}
	}
	
	public List<Edge> getOutgoing() {
		if (isPalindrome) {
			return outgoing;
		} else {
			List<Edge> tmpOutgoing = new LinkedList<Edge>();
			for (Edge oEdge: outgoing) {
				if (oEdge.getStrand() == 0) {
					tmpOutgoing.add(oEdge);
				}
			}
			for (Edge iEdge: incoming) {
				if (iEdge.getStrand() == 1) {
					tmpOutgoing.add(iEdge);
				}
			}
			return tmpOutgoing;
		}
	}
	
	public void addIncoming(Edge e, Byte pole) {
		if (e.getStrand() == 0) {
			if (!incoming.contains(e)) {
				incoming.add(e);
				incomingPoles.add(pole);
			}
		} else {
			// store info for the edge's positive strand orientation
			if (!outgoing.contains(e)) {
				outgoing.add(e);
				if (isPalindrome) {
					outgoingPoles.add(pole);
				} else {
					outgoingPoles.add(getFlippedPole(pole));
				}
			}
		}
	}
	
	public void addOutgoing(Edge e, Byte pole) {
		if (e.getStrand() == 0) {
			if (!outgoing.contains(e)) {
				outgoing.add(e);
				outgoingPoles.add(pole);
			}
		} else {
			// store info for the edge's positive strand orientation
			if (!incoming.contains(e)) {
				incoming.add(e);
				if (isPalindrome) {
					incomingPoles.add(pole);
				} else {
					incomingPoles.add(getFlippedPole(pole));
				}
			}
		}
	}
	
	/**
	 * Determines the pole for edge e in requested direction
	 * 
	 * @param e edge of interest
	 * @param type direction ("in" or "out")
	 * @return pole (0 if edge connects at top of vertex,
	 *               1 if edge connects at bottom of verex,
	 *              -1 if edge does not connect to this vertex)
	 */
	public byte getPole(Edge e, String direction) {
		if (direction.equals("in")) {
			return getIncomingPole(e);
		} else if (direction.equals("out")) {
			return getOutgoingPole(e);
		} else {
			throw new IllegalArgumentException("Vertex Error: Unknown direction " + direction);
		}
	}
	
	private byte getIncomingPole(Edge e) {
		byte pole = -1;
		if (e.getStrand() == 0) {
			int ii = incoming.indexOf(e);
			if (ii != -1) {
				// e is present in the incoming list
				if (isPalindrome) {
					// guarantee pole=0 if palindrome
					pole = 0;
				} else {
					pole = incomingPoles.get(ii);
				}
			}
		} else {
			// want the rc of stored values
			int oi = outgoing.indexOf(e);
			if (oi != -1) {
				// e is present in the outgoing list
				if (isPalindrome) {
					// guarantee pole=0 if palindrome
					pole = 0;
				} else {
					pole = outgoingPoles.get(oi);
					// adjust for opposite direction
					pole = getFlippedPole(pole);
				}
			}
		}
		return pole;
	}
	
	private byte getOutgoingPole(Edge e) {
		byte pole = -1;
		if (e.getStrand() == 0) {
			int oi = outgoing.indexOf(e);
			if (oi != -1) {
				// e is present in the outgoing list
				if (isPalindrome) {
					// guarantee pole=0 if palindrome
					pole = 0;
				} else {
					pole = outgoingPoles.get(oi);
				}
			}
		} else {
			// want the rc of stored values
			int ii = incoming.indexOf(e);
			if (ii != -1) {
				// e is present in the incoming list
				if (isPalindrome) {
					// guarantee pole=0 if palindrome
					pole = 0;
				} else {
					pole = incomingPoles.get(ii);
					// adjust for opposite direction
					pole = getFlippedPole(pole);
				}
			}
		}
		return pole;
	}
	
	public byte getFlippedPole(byte pole) {
		if (isPalindrome) {
			return 0;
		} else if (pole == 0) {
			return 1;
		} else {
			return 0;
		}
	}
	
}
